# ------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ------------------------------------------------------------------------------

import os
from tqdm import tqdm
from pywebhdfs import errors
from mssqlctl.cli.commands.hdfs.util import FormatWebHDFS
from knack.prompting import (prompt_pass, NoTTYException)
from mssqlctl.cli.core.util import raise_command_error
from mssqlctl.cli.commands.hdfs.hdfsshell import HDFSShell
from mssqlctl.cli.commands.hdfs.exceptions import HdfsError

"""
NOTE: The BDC authentication has a problem in that each endpoint uses an inconsistent user and password scheme.
      The controller takes a user and password.  SQL takes a password and hard coded user 'sa'.  And HDFS takes
      a password and hard coded user 'root'.  This isn't usable and doesn't match the experience when
      Kerberos and Active Directory are supported.  So for now the password for HDFS is the controller password.
      A future change will remove the three current identities and move to a cluster user/password with
      permissions determining the endpoint behavior.  For this reason the controller password needs to be
      the same across all endpoints in the short term and the controller password is retrieved below.  
"""


def _get_knox_username():
    # Note for now the knox username is hard coded to 'root'.
    return "root"


def _get_knox_password(client):
    # Note for now if the environment variable isn't set prompt for the password..
    if not os.getenv('KNOX_PASSWORD'):
        if not os.getenv('CONTROLLER_PASSWORD'):
            try:
                client.stdout(
                    "The Knox or controller password is needed.  It can be set for the session by setting the environment variable KNOX_PASSWORD.")
                client.stdout("If Knox password is not set then CONTROLLER_PASSWORD is used.")
                password = prompt_pass('Enter Knox Password: ')
            except NoTTYException:
                raise NoTTYException('Missing KNOX_PASSWORD or CONTROLLER_PASSWORD environment variable in non-interactive environment.')
        else:
            password = os.environ['CONTROLLER_PASSWORD']
    else:
        password = os.environ['KNOX_PASSWORD']

    return password


class Progress(object):
    def __init__(self, client):
        self.progress_bar = None
        self.client = client

    def initialize(self, description, total_size):
        if self.progress_bar:
            self.progress_bar.close()

        self.description = description
        self.total_size = total_size
        self.progress_bar = tqdm(desc=description, total=total_size, ascii=".oO0", unit_scale=True, dynamic_ncols=True)

    def update(self, current_position):
        self.progress_bar.update(current_position)


def _login(client):
    endpoint = client.get_knox_endpoint()
    username = _get_knox_username()
    password = _get_knox_password(client)

    client.login(endpoint, username, password)
    return


def hdfs_shell(client):
    endpoint = client.get_knox_endpoint()
    username = _get_knox_username()
    password = _get_knox_password(client)

    hdfs = HDFSShell(client, endpoint, username, password)
    hdfs.cmdloop()
    return


def hdfs_ls(client, path):
    try:
        _login(client)
        result = client.ls(path)
        output = FormatWebHDFS.convert_filestatuses_to_list(result["FileStatuses"])
    except errors.FileNotFound as e:
        raise HdfsError("ls: No such file or directory\n")

    return output


def hdfs_exists(client, path):
    _login(client)
    result = client.exists(path)
    return result


def hdfs_mkdir(client, path):
    _login(client)
    result = client.mkdir(path)
    return result


def hdfs_mv(client, source_path, target_path):
    _login(client)
    result = client.mv(source_path, target_path)
    return result


def hdfs_create(client, path, data):
    _login(client)
    result = client.create_file(path, data)
    return result


def hdfs_cat(client, path, offset, length):
    try:
        _login(client)
        output = client.cat_file(path)
    except errors.FileNotFound as e:
        raise HdfsError("cat: No such file or directory\n")

    return output


def hdfs_rm(client, path):
    try:
        _login(client)
        output = client.remove_file(path, False)
    except errors.FileNotFound as e:
        raise HdfsError("rm: No such file or directory\n")

    return output


def hdfs_rmr(client, path):
    try:
        _login(client)
        output = client.remove_file(path, True)
    except errors.FileNotFound as e:
        raise HdfsError("rmr: No such file or directory\n")

    return output


def hdfs_chmod(client, path, permission):
    try:
        _login(client)
        output = client.set_permission(path, permission)
    except errors.FileNotFound as e:
        raise HdfsError("chmod: No such file or directory\n")

    return output


def hdfs_chown(client, path, owner, group):
    try:
        _login(client)
        output = client.set_owner(path, owner, group)
    except errors.FileNotFound as e:
        output = "chown: No such file or directory\n"

    return output


def hdfs_acl_modify(client, path, aclspec):
    try:
        _login(client)
        output = client.modify_acl_entries(path, aclspec)
    except errors.FileNotFound as e:
        raise HdfsError("acl modify: No such file or directory\n")

    return output


def hdfs_acl_remove(client, path, aclspec=None):
    try:
        _login(client)
        if aclspec is None:
            output = client.remove_acl(path)
        elif aclspec.lower() == 'default':
            output = client.remove_default_acl_entries(path)
        else:
            output = client.remove_acl_entries(path, aclspec)
    except errors.FileNotFound as e:
        raise HdfsError("acl remove: No such file or directory\n")

    return output


def hdfs_acl_set(client, path, aclspec):
    try:
        _login(client)
        output = client.set_acl(path, aclspec)
    except errors.FileNotFound as e:
        raise HdfsError("acl set: No such file or directory\n")

    return output


def hdfs_acl_status(client, path):
    try:
        _login(client)
        output = client.get_acl_status(path)
    except errors.FileNotFound as e:
        raise HdfsError("acl status: No such file or directory\n")

    return output


def hdfs_xattr_get(client, path, name):
    try:
        _login(client)
        output = client.get_xattr(path, name)
    except errors.FileNotFound as e:
        raise HdfsError("xattr get: No such file or directory\n")

    return output


def hdfs_xattr_set(client, path, name, value):
    try:
        _login(client)
        output = client.set_xattr(path, name, value)
    except errors.FileNotFound as e:
        raise HdfsError("xattr set: No such file or directory\n")

    return output


def hdfs_xattr_modify(client, path, name, value):
    try:
        _login(client)
        output = client.modify_xattr(path, name, value)
    except errors.FileNotFound as e:
        raise HdfsError("xattr modify: No such file or directory\n")

    return output


def hdfs_xattr_remove(client, path, name):
    try:
        _login(client)
        output = client.delete_xattr(path, name)
    except errors.FileNotFound as e:
        raise HdfsError("xattr remove: No such file or directory\n")

    return output


def hdfs_xattr_status(client, path):
    try:
        _login(client)
        output = client.get_xattr_status(path)
    except errors.FileNotFound as e:
        raise HdfsError("xattr status: No such file or directory\n")

    return output


def hdfs_cp_l2r(client, local_path, remote_path):
    try:
        _login(client)
        progress_bar = Progress(client)
        output = client.copy_local_to_remote(local_path, remote_path, progress_bar)
    except errors.FileNotFound as e:
        raise HdfsError("copy local to remote status: No such file or directory\n")

    return output


def hdfs_cp_r2l(client, remote_path, local_path):
    try:
        _login(client)
        progress_bar = Progress(client)
        output = client.copy_remote_to_local(remote_path, local_path, progress_bar)
    except errors.FileNotFound as e:
        raise HdfsError("copy remote to local status: No such file or directory\n")

    return output


def hdfs_mount_create(client, remote_uri, mount_path):
    """
    Method to create mounts using HDFS tiering. Invoked via:
        mssqlctl storage mount create --remote-uri <..>
                                      --mount-path <..>
    Credentials for creating the mount can be provided with the environment
    varaible, MOUNT_CREDENTIALS.

    :param remote_uri: URI of the remote store to mount.
    :param mount_path: HDFS path where the mount has to be created.
    """
    try:
        if not all([remote_uri, mount_path]):
            raise ValueError('Missing arguments [--remote-uri, --mount-path] '
                             'are required.')

        credentials_json = None
        if not os.getenv('MOUNT_CREDENTIALS'):
            client.stdout('Warning: Attempting to mount {0} without using any '
                          'credentials.'.format(remote_uri))
        else:
            from mssqlctl.cli.commands.hdfs.util import get_credentials_json
            credentials_json = get_credentials_json(
                os.environ['MOUNT_CREDENTIALS'])

        client.apis.controller.create_mount(remote_uri, mount_path, credentials_json)

        client.stdout('Create mount for {0} submitted successfully. Check '
                      'mount status for progress'.format(mount_path))
    except Exception as ex:
        raise_command_error(ex)


def hdfs_mount_delete(client, mount_path):
    """
    Method to delete mounts created used HDFS tiering.
    :param client: API client to use
    :param mount_path: HDFS path corresponding to the mount to be deleted.
    :return:
    """
    try:
        if not mount_path:
            raise ValueError('Missing argument [--mount-path] is required.')

        client.apis.controller.delete_mount(mount_path)

        client.stdout('Delete mount for {0} submitted successfully. Check '
                      'mount status for progress'.format(mount_path))
    except Exception as ex:
        raise_command_error(ex)


def hdfs_mount_status(client, mount_path=None):
    """
    Method to get info on mounts.
    :param client: API client to use
    :param mount_path: HDFS path corresponding to the mount.
    :return:
    """
    try:
        # if path is not specified, just list all mounts.
        return client.apis.controller.get_mount_status(mount_path)
    except Exception as ex:
        raise_command_error(ex)


def hdfs_mount_refresh(client, mount_path):
    """
    Method to refresh mounts created used HDFS tiering.
    :param client: API client to use
    :param mount_path: HDFS path corresponding to the mount to be refreshed.
    :return:
    """
    try:
        if not mount_path:
            raise ValueError('Missing argument [--mount-path] is required.')

        client.apis.controller.refresh_mount(mount_path)

        client.stdout('Refresh mount for {0} submitted successfully. Check '
                      'mount status for progress'.format(mount_path))
    except Exception as ex:
        raise_command_error(ex)
