# ------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ------------------------------------------------------------------------------

from knack.prompting import (prompt, prompt_pass, NoTTYException)
from mssqlctl.cli.core.util import (load_kube_config, get_controller_endpoint, raise_command_error, eula_prompt)
from mssqlctl.cli.core.exceptions import (ApiError, CliError, HTTP_UNAUTHORIZED)

import os


def login(client, cluster_name=None, controller_username=None, controller_endpoint=None, accept_eula=None):
    """
    ```sh
    mssqlctl login
    ```
    """
    try:
        stdout = client.stdout

        # -- AD/Basic: prompt for username --
        if not controller_username:
            if not os.getenv('CONTROLLER_USERNAME'):
                try:
                    controller_username = prompt('Controller Username: ')
                except NoTTYException:
                    raise NoTTYException('Missing --controller-username arg in non-interactive environment.')
            else:
                controller_username = os.environ['CONTROLLER_USERNAME']

        # -- AD/Basic: prompt for password --
        if not os.getenv('CONTROLLER_PASSWORD'):
            try:
                controller_password = prompt_pass('Controller Password: ')
            except NoTTYException:
                raise NoTTYException('Missing CONTROLLER_PASSWORD environment variable in non-interactive environment.')
        else:
            controller_password = os.environ['CONTROLLER_PASSWORD']

        # -- Endpoint: Check Kube Config --
        if not controller_endpoint:
            # -- AD/Basic: prompt for namespace --
            if not cluster_name:
                try:
                    cluster_name = prompt('Cluster Name: ')
                except NoTTYException:
                    raise NoTTYException('When not specifying the --endpoint and using the kube config, please specify '
                                         'the cluster name as an argument via --cluster-name/-n.')
            # Attempt to load the kube config if it exists
            if os.environ.get('KUBECONFIG', None) or os.environ.get('KUBECTL_CONTEXT', None) \
                    or os.path.isfile(os.path.expanduser('~/.kube/config')):
                try:
                    # Attempt to get endpoint via kube config
                    load_kube_config(os.getenv('KUBECTL_CONTEXT', None))
                    controller_endpoint = get_controller_endpoint(cluster_name)
                except Exception as e:
                    raise CliError('Please check your kube config or specify the correct controller endpoint with: '
                                   '--controller-endpoint https://<ip>:<port>. {0}'.format(e))
            # -- Endpoint: Check Profile --
            if not controller_endpoint:
                controller_endpoint = client.controller_endpoint()
            # -- Endpoint: Prompt --
            if not controller_endpoint:
                try:
                    controller_endpoint = prompt('Controller Endpoint: ')
                except NoTTYException:
                    raise NoTTYException('Missing --controller-endpoint arg in non-interactive environment. '
                                         'You may also place your kube config in the default location of '
                                         '<user home>/.kube/config or use the KUBECONFIG env var to specify the '
                                         'location of the config, i.e. export KUBECONFIG=path/to/.kube/config.')

        # -- enforce EULA --
        accept_eula = eula_prompt(client, accept_eula, True)

        client.login(controller_endpoint, controller_username, controller_password, cluster_name, accept_eula.lower())
    except ApiError as ex:
        if ex.status == HTTP_UNAUTHORIZED:
            raise CliError('Wrong username or password.')
        else:
            raise CliError(ex)
    except NoTTYException as e:
        raise CliError('{0} Please specify arguments and/or environment variables in non-interactive mode. '
                       'For a better understanding of the arguments, please run \'mssqlctl login --help\'.'.format(e))
    except Exception as ex:
        raise CliError(ex)

    client.stdout('Logged in successfully to `{0}`'.format(controller_endpoint))


def logout(client):
    """
    ```sh
    mssqlctl logout
    ```
    """
    try:
        client.logout()
    except Exception as e:
        raise_command_error(e)
