# ------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ------------------------------------------------------------------------------

"""
Client for all CLI actions.
"""

from mssqlctl.cli.core.output import OutputStream
from mssqlctl.cli.core.profile import MSSQLProfile
from mssqlctl.cli.core.clients.process_client import Command as ProcessClient
from mssqlctl.cli.core.clients.controller_client import ControllerClient
from mssqlctl.cli.core.clients.control_deploy_client import ControlDeployClient
from mssqlctl.cli.core.exceptions import CliError
from mssqlctl.cli.core.configuration import Configuration
from mssqlctl.cli.core.util import eula_prompt
from abc import ABCMeta
from six import add_metaclass

__all__ = ['client']


def beget_cli_client(check_auth=True, check_eula=True):
    """
    The factory function used to apply the common `CliClient` to a custom
    commands's command group.

    :param check_auth: Check for authentication on this command group.
    :param check_eula: Check for eula acceptance on this command group.
    :return: A function.
    """
    def beget(_):
        return CliClient(check_auth=check_auth, check_eula=check_eula)

    return beget

client = beget_cli_client  # Export

# ============================================================================ #
# ============================================================================ #
# ============================================================================ #

@add_metaclass(ABCMeta)
class BaseCliClient(object):
    def __init__(self, check_auth=True, check_eula=True):
        self.profile = MSSQLProfile()

        if check_auth:
            access_token = self.profile.access_token
            controller_endpoint = self.profile.controller_endpoint
            cluster_name = self.profile.cluster_name

            if not controller_endpoint or not access_token:
                raise CliError(
                    'Please run \'{0} login\' to first authenticate.'.format(Configuration().CLI_NAME))

        if check_eula:
            accept_eula = self.profile.accept_eula
            if not accept_eula:
                self.profile.accept_eula = eula_prompt(self)

    @property
    def profile(self):
        return self._profile

    @profile.setter
    def profile(self, profile):
        self._profile = profile

    @property
    def stdout(self):
        return OutputStream().stdout.write

    @property
    def stderr(self):
        return OutputStream().stderr.write

# ============================================================================ #
# ============================================================================ #
# ============================================================================ #


class Apis(object):
    def __init__(self, controller, control_plane):
        self._controller = controller
        self._control_deploy = control_plane

    @property
    def control_deploy(self):
        """
        Gets the reference to the Control Plane deployment client
        :return:
        """
        return self._control_deploy

    @property
    def controller(self):
        """
        Gets the reference to the Controller API client
        :return:
        """
        return self._controller

# ============================================================================ #
# ============================================================================ #
# ============================================================================ #


class CliClient(BaseCliClient):
    """
    Default client injected in every command group. For further command
    customization extend this class.
    """
    def __init__(self, check_auth=True, check_eula=True):
        super(CliClient, self).__init__(check_auth, check_eula)

        self._apis = Apis(
            ControllerClient(
                self.profile.controller_endpoint,
                access_token=self.profile.access_token,
                namespace=self.profile.cluster_name
            ),
            ControlDeployClient(self.profile.cluster_name)
        )

        self._process = ProcessClient

    @property
    def apis(self):
        """
        Gets the reference to the client apis
        """
        return self._apis

    @property
    def process(self):
        """
        Gets the reference to the process client.
        """
        return self._process

    def import_api(self, api_module):
        """
        Gets the instance of the API class from the given `api_module`. This is
        a common pattern for instantiated APIs given a API module name.

        :param api_module: The name of API module.
        :return: The instance of the API.
        """
        from importlib import import_module
        from inspect import (getmembers, isclass)

        mod = import_module('mssqlctl.cli.core.api.{}'.format(api_module))
        name, klass = getmembers(mod, isclass)[1]
        return klass(api_client=self.apis.controller.api_client)
