# ------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ------------------------------------------------------------------------------

from mssqlctl.cli.core.configuration import Configuration
from mssqlctl.cli.core.exceptions import ApiError, OPERATION_TIMED_OUT, CONNECTION_REFUSED
from mssqlctl.cli.core.clients.api_client import ApiClient
from mssqlctl.cli.core.api.app_router_api import AppRouterApi
from mssqlctl.cli.core.api.cluster_router_api import ClusterRouterApi
from mssqlctl.cli.core.api.token_router_api import TokenRouterApi
from mssqlctl.cli.core.exceptions import (ControllerError, http_status_codes)
from mssqlctl.cli.core.logging import get_logger
from urllib3.exceptions import (NewConnectionError, MaxRetryError)
import uuid
import json

logger = get_logger(__name__)


class ControllerClient(object):
    """
    The client for managing HTTP API calls to controller-plane.
    """
    def __init__(self, endpoint, access_token=None, username=None,
                 password=None, namespace=None):

        config = Configuration()
        config.host = endpoint

        self._init_api_client(config, access_token, username, password)
        self._api = ClusterRouterApi(api_client=self.api_client)

    def _init_api_client(self, config, access_token=None, username=None, password=None):
        self._api_client = ApiClient(
            configuration=config,
            header_name='version',
            header_value=config.CLUSTER_VERSION
        )

        # -- Apply Bearer access-token all HTTP requests --
        if access_token:
            self._api_client.set_default_header(
                'Authorization',
                'Bearer {}'.format(access_token)
            )
        elif username and password:
            config.username = username
            config.password = password
            self._api_client.set_default_header(
                'Authorization',
                '{}'.format(self.api_client.configuration.get_basic_auth_token())
            )

        self._api = ClusterRouterApi(api_client=self.api_client)

    @property
    def api_client(self):
        """
        Gets the internal swagger client reference.
        """
        return self._api_client

    def request(self, resource_path, method,
                path_params=None, query_params=None, header_params=None,
                body=None, post_params=None, files=None,
                response_type=None, auth_settings=None,
                return_http_data_only=None, collection_formats=None,
                preload_content=True, request_timeout=None):

        """
        Makes the HTTP request (synchronous) and returns deserialized data.

        :param resource_path: Path to method endpoint.
        :param method: Method to call.
        :param path_params: Path parameters in the url.
        :param query_params: Query parameters in the url.
        :param header_params: Parameters to be placed in the request header.
        :param body: body: Request body.
        :param post_params: Request post form parameters, for
                            `application/x-www-form-urlencoded`,
                            `multipart/form-data`.
        :param files: key -> filename, value -> filepath, for
                      `multipart/form-data`.
        :param response_type: Response data type.
        :param auth_settings: Auth Settings names for the request.
        :param return_http_data_only: response data without head status code
                                      and headers
        :param collection_formats: dict of collection formats for path, query,
                                   header, and post parameters.
        :param preload_content: if False, the urllib3.HTTPResponse object will
                                be returned without reading/decoding response
                                data. Default is True.
        :param request_timeout: timeout setting for this request. If one
                                number provided, it will be total request
                                timeout. It can also be a pair (tuple) of
                                (connection, read) timeouts.
        :return: The response directly.
        """

        return self._api_client.call_api(
            resource_path,
            method,
            path_params=path_params,
            query_params=query_params,
            header_params=header_params,
            body=body,
            post_params=post_params,
            files=files,
            response_type=response_type if response_type else object,
            auth_settings=auth_settings if auth_settings else [],
            _return_http_data_only=return_http_data_only,
            collection_formats=collection_formats,
            _preload_content=preload_content,
            _request_timeout=request_timeout
        )

    # ------------------------------------------------------------------------ #
    # token router: /
    # ------------------------------------------------------------------------ #

    def acquire_token_with_basic(self, endpoint, username, password):
        def _flush_config_on_error():
            default_config = Configuration()
            default_config.username = ''
            default_config.password = ''

        try:
            config = Configuration()
            config.username = username
            config.password = password
            config.host = endpoint

            # Call an API to check for HTTP 401 against basic un:pw
            client = ApiClient(configuration=config)
            api = TokenRouterApi(api_client=client)
            response = api.token_post()

            # -- update access token for all HTTP requests --
            access_token = response.access_token
            self._init_api_client(config, access_token)

            return {'accessToken': access_token}
        except ApiError as e:
            _flush_config_on_error()
            raise ApiError(status=e.status, reason=e.reason)
        except Exception as e:
            _flush_config_on_error()
            raise ApiError(e)

    # ------------------------------------------------------------------------ #
    # app router: /app
    # ------------------------------------------------------------------------ #

    def create_app(self, spec, archive):
        try:
            api = AppRouterApi(api_client=self._api_client)
            return api.create_app(spec=spec, package=archive)
        except ApiError as e:
            raise ControllerError(e, status=e.status)

    def update_app(self, spec, archive):
        try:
            api = AppRouterApi(api_client=self._api_client)
            return api.update_app(spec=spec, package=archive)
        except ApiError as e:
            raise ControllerError(e, status=e.status)

    def delete_app(self, name, version=None):
        try:
            api = AppRouterApi(api_client=self._api_client)
            return api.app_by_name_by_version_delete(name, version)
        except ApiError as e:
            raise ControllerError(e, status=e.status)

    def list_apps(self, name=None):
        try:
            api = AppRouterApi(api_client=self._api_client)
            if not name:
                res = api.app_get()
            else:
                res = api.app_by_name_get(name=name)
            apps = []
            for app in res:
                apps.append(app.to_dict())
            return apps
        except ApiError as e:
            if e.status == http_status_codes.not_found:
                return []
            else:
                raise ControllerError(e, status=e.status)

    def list_app(self, name, version=None):
        try:
            api = AppRouterApi(api_client=self._api_client)
            return [api.app_by_name_by_version_get(name, version).to_dict()]
        except ApiError as e:
            if e.status == http_status_codes.not_found:
                return []
            else:
                raise ControllerError(e, status=e.status)

    # ------------------------------------------------------------------------ #
    # cluster Router (storage): /cluster
    # ------------------------------------------------------------------------ #

    def create_mount(self, remote_uri, local_path, credentials):
        """
        Create a mount.
        :param remote_uri: URI of the store to mount.
        :param local_path: Local HDFS path which will serve as the mount point.
        :param credentials: the credentials to create the mount as a JSON.
        """

        # generate a request id
        request_id = str(uuid.uuid4())

        try:
            # Create a mount
            logger.debug(
                'Creating mount %s for remote uri %s.' %
                (local_path, remote_uri)
            )

            return self._api.create_mount(
                request_id,
                'close',
                remote_uri,
                local_path,
                credentials=credentials
            )
        except ApiError as e:
            status = e.status
            if status == http_status_codes.bad_gateway:
                msg = '%d - Bad Gateway Error.' % (status)
                logger.error(msg)
            else:
                # Log the reason if it was included
                msg = 'Failed to create mount with status code [%d]' % status
                try:
                    msg = '%s. Reason: %s' % (msg, json.loads(e.body)['reason'])
                except (ValueError, TypeError):
                    pass

                logger.error(msg)

            raise ControllerError(Exception(msg), status=status)
        except Exception as e:
            raise ControllerError(e)

    def delete_mount(self, local_path):
        """
        Delete a mount.
        :param local_path: mount point to delete
        """

        request_id = str(uuid.uuid4())
        try:
            # Delete the mount
            logger.debug('Deleting mount %s.' % local_path)
            return self._api.delete_mount(request_id, 'close', local_path)
        except ApiError as e:
            status = e.status
            if status == http_status_codes.bad_gateway:
                msg = '%d - Bad Gateway Error.' % (status)
                logger.error(msg)
                return False
            else:
                # Log the reason if it was included
                msg = 'Failed to delete mount with status code [%d]' % status
                try:
                    msg = '%s. Reason: %s' % (msg, json.loads(e.body)['reason'])
                except (ValueError, TypeError):
                    pass

                logger.error(msg)

            raise ControllerError(Exception(msg), status=status)
        except Exception as e:
            raise ControllerError(e)

    def refresh_mount(self, local_path):
        """
        Refresh a mount.
        :param local_path: mount point to refresh
        """
        request_id = str(uuid.uuid4())
        try:
            # Delete the mount
            logger.debug('Refreshing mount %s.' % local_path)
            return self._api.refresh_mount(request_id, 'close', local_path)
        except ApiError as e:
            status = e.status
            if status == http_status_codes.bad_gateway:
                msg = '%d - Bad Gateway Error.' % (status)
            else:
                # Log the reason if it was included
                msg = 'Failed to refresh mount with status code [%d]. Reason %s.' % \
                      (status, e.reason)

            logger.error(msg)
            raise ControllerError(Exception(msg), status=status)
        except Exception as e:
            raise ControllerError(e)

    def get_mount_status(self, local_path):
        """
        Get the status of mount(s).
        :param local_path: mount path
        :return: A Json for the status of a specified mount. If local_path is
        None, returns the status of all existing mounts.
        """
        request_id = str(uuid.uuid4())
        try:
            # Delete the mount
            if local_path is None:
                logger.debug('Getting status of all mounts.')
                return self._api.get_mounts(request_id, 'close')
            else:
                logger.debug('Getting status of mount %s.' % local_path)
                return self._api.get_mounts(
                    request_id, 'close', mount=local_path)
        except ApiError as e:
            status = e.status
            if status == http_status_codes.bad_gateway:
                logger.error('%d - Bad Gateway Error.' % (status))
                return False
            else:
                # Log the reason if it was included
                if local_path is None:
                    msg = 'Failed to get status of mounts with status code ' \
                          '[%d]' % status
                else:
                    msg = 'Failed to get status of mount %s with status code ' \
                          '[%d]' % (local_path, status)
                try:
                    msg = '%s. Reason: %s' % (msg, e.body)
                except ValueError:
                    pass

                logger.error(msg)
                raise ControllerError(Exception(msg), status=status)
        except Exception as e:
            raise ControllerError(e)

    # ------------------------------------------------------------------------ #
    # cluster Router: /cluster
    # ------------------------------------------------------------------------ #

    def create_cluster(self, cluster_config):
        """
        Create cluster
        """

        # Generate a request id
        request_id = str(uuid.uuid4())
        try:
            self._api.create_cluster(
                x_request_id=request_id,
                connection='close',
                data=json.loads(cluster_config)
            )
            return True
        except ApiError as e:
            if e.status == http_status_codes.bad_gateway:
                logger.debug('%d - Bad Gateway Error. Retrying ...' % (e.status))
                return False
            else:
                logger.error('Failed to create cluster '
                             'with status code [%d]' % (e.status))
                # Log the reason if it was included
                try:
                    logger.error('Reason: %s' % (json.loads(e.body)['reason']))
                except (ValueError, TypeError):
                    pass
                logger.debug(e)
                msg = 'Failed to create cluster with status ' \
                      'code [%d]' % (e.status)
                raise ControllerError(Exception(msg), status=e.status)
        except Exception as e:
            logger.debug(e)
            return False

    def delete_cluster(self):
        """
        Delete cluster
        """

        # Generate a request id
        request_id = str(uuid.uuid4())
        try:
            self._api.delete_cluster(
                x_request_id=request_id,
                connection='close'
            )
        except ApiError as e:
            msg = 'Failed to delete cluster with status code ' \
                  '[%d]' % (e.status)
            logger.error(msg)
            raise ControllerError(Exception(msg), status=e.status)

    def upgrade_control_plane(self, target_version):
        """
        Upgrade control plane
        """

        # Generate a request id
        request_id = str(uuid.uuid4())
        try:
            self._api.upgrade_control_plane(
                x_request_id=request_id,
                connection='close',
                data=target_version
            )
            return True
        except ApiError as e:
            if e.status == http_status_codes.bad_gateway:
                logger.debug('%d - Bad Gateway Error. Retrying ...' % (e.status))
                return False
            else:
                logger.error('Failed to upgrade control plane '
                             'with status code [%d]' % (e.status))
                # Log the reason if it was included
                try:
                    logger.error('Reason: %s' % (json.loads(e.body)['reason']))
                except (ValueError, TypeError):
                    pass

                msg = 'Failed to upgrade control plane with status ' \
                      'code [%d]' % (e.status)
                raise ControllerError(Exception(msg), status=e.status)
        except Exception as e:
            logger.debug(e)
            return False

    def upgrade_bdc(self, target_version):
        """
        Upgrade Big Data Cluster
        """

        # Generate a request id
        request_id = str(uuid.uuid4())
        try:
            self._api.upgrade_bdc(
                x_request_id=request_id,
                connection='close',
                data=target_version
            )
            return True
        except ApiError as e:
            if e.status == http_status_codes.bad_gateway:
                logger.debug('%d - Bad Gateway Error. Retrying ...' % (e.status))
                return False
            else:
                logger.error('Failed to upgrade cluster '
                             'with status code [%d]' % (e.status))
                # Log the reason if it was included
                try:
                    logger.error('Reason: %s' % (json.loads(e.body)['reason']))
                except (ValueError, TypeError):
                    pass

                msg = 'Failed to upgrade cluster with status ' \
                      'code [%d]' % (e.status)
                raise ControllerError(Exception(msg), status=e.status)
        except Exception as e:
            logger.debug(e)
            return False

    def get_controller_logs(self, offset):
        """
        Streams the progress of cluster creation
        """

        try:
            return self._api.get_logs(
                x_request_id=str(uuid.uuid4()),
                connection='close',
                offset=offset,
                _request_timeout=60
            )
        except Exception as e:
            logger.debug(e)
            return None

    def get_cluster(self):
        """
        Get cluster config
        """

        # Generate a request id
        request_id = str(uuid.uuid4())
        try:
            response = self._api.get_cluster(
                x_request_id=request_id,
                connection='close',
                _request_timeout=30
            )
            return response['spec']
        except ApiError as e:
            raise ControllerError(e, status=e.status)

    def get_endpoints(self, endpoint_name=None):
        try:
            if not endpoint_name:
                return self._api.endpoints_get()
            else:
                return self._api.endpoints_by_name_get(endpoint_name)
        except ApiError as e:
            raise ControllerError(e, status=e.status)

    def get_cluster_state(self, loggingEnabled=True):
        """
        Get cluster state
        """
        # Generate a request id
        request_id = str(uuid.uuid4())
        try:
            result = self._api.get_bdc_status(
                request_id,
                'close',
                _request_timeout=30
            )

            return result['state']
        except ApiError as e:
            if (loggingEnabled):
                if e.status == http_status_codes.not_found:
                    logger.info('Cannot find any cluster')
                elif e.status == http_status_codes.unauthorized:
                    logger.info('Failed to get the cluster information. Invalid '
                                'username or password')
                elif e.status == 0:
                    logger.info(e.reason)
                elif e.status == OPERATION_TIMED_OUT:
                    logger.debug('Failed to get cluster status '
                                'with code [%d]' % (e.status))
                else:
                    logger.info('Failed to get cluster status '
                                'with code [%d]' % (e.status))
                    try:
                        logger.info('Reason: %s' % (json.loads(e.body)['reason']))
                    except (ValueError, TypeError):
                        pass
            logger.debug(e)
            return None
        except Exception as e:
            logger.debug(e)
            return None

    def get_bdc_status(self, service_name=None, resource_name=None, loggingEnabled=True, **kwargs):
        """
        Get BDC status
        """
        try:
            request_id = str(uuid.uuid4())
            conn = 'close'
            time = 30
            api = ClusterRouterApi(api_client=self._api_client)
            if service_name:
                if resource_name:
                    return api.get_bdc_service_resource_status(
                        request_id,
                        conn,
                        service_name,
                        resource_name,
                        **kwargs,
                        _request_timeout=time
                    )
                else:
                    return api.get_bdc_service_status(
                        request_id,
                        conn,
                        service_name,
                        **kwargs,
                        _request_timeout=time
                    )
            elif resource_name:
                return api.get_bdc_resource_status(
                    request_id,
                    conn,
                    resource_name,
                    **kwargs,
                    _request_timeout=time
                )
            else:
                return api.get_bdc_status(
                    request_id,
                    conn,
                    **kwargs,
                    _request_timeout=time
                )
        except ApiError as e:
            raise ControllerError(e, status=e.status)

    def get_control_status(self, control_plane_state=False, loggingEnabled = True):
        """
        Get control plane status
        """
        # Generate a request id
        request_id = str(uuid.uuid4())
        try:
            result = self._api.get_control_status(
                x_request_id=request_id,
                connection='close',
                _request_timeout=30
            )

            if not control_plane_state:
                return result
            else:
                return result['state']
        except ApiError as e:
            raise ControllerError(e, status=e.status)

    def is_ready(self, loggingEnabled=True):
        """
        Check if controller is ready
        """

        request_id = str(uuid.uuid4())
        try:
            if loggingEnabled:
                logger.debug('Sending Get request.')

            self._api.get_home(
                x_request_id=request_id,
                connection='close',
                _request_timeout=30
            )
            return True
        except ApiError as e:
            if e.status not in [http_status_codes.bad_gateway, http_status_codes.service_unavailable, OPERATION_TIMED_OUT, CONNECTION_REFUSED] :
                if loggingEnabled:
                    logger.info('Failed to retrieve home with '
                                 'status code [%d]' % (e.status))
                try:
                    logger.error('Reason: %s' % (json.loads(e.body)['reason']))
                except ValueError:
                    pass
                msg = 'Failed to retrieve home with status ' \
                      'code [%d]' % (e.status)
                logger.debug(e)
                raise ControllerError(Exception(msg), status=e.status)
            return False
        except (NewConnectionError, MaxRetryError, Exception) as e:
            return False

    def update_password(self, service_name, service_username,
                        new_password):
        """
        Update service password
        """

        request_id = str(uuid.uuid4())

        try:
            data = {}
            data['password'] = new_password
            return self._api.update_password(
                x_request_id=request_id,
                connection='close',
                service_name=service_name,
                service_username=service_username,
                data=data
            )
        except ApiError as e:
            logger.error('Failed to set password for '
                         'service [%s] and user [%s] with status code '
                         '[%d]' % (service_name, service_username,
                                   e.status))
            try:
                logger.error('Reason: %s' % (json.loads(e.body)['reason']))
            except (ValueError, TypeError):
                pass
            raise Exception(
                'Failed to set password for service "%s" user "%s".' % (
                    service_name, service_username))
