# ------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ------------------------------------------------------------------------------

from __future__ import absolute_import
from mssqlctl.cli.core.util import FileUtil
from mssqlctl.cli.core.configuration import Configuration
from mssqlctl.cli.core.exceptions import CliError
from mssqlctl.cli.core.logging import get_logger
from knack.util import ensure_dir

import errno
import json
import os

log = get_logger(__name__)


class MSSQLProfile(object):

    def __init__(self, auth_token=None,
                 controller_endpoint=None,
                 controller_username=None,
                 cluster_name=None,
                 accept_eula=None):
        self._profile_cache = _GLOBAL_MSSQL_PROFILE_CACHE

        if auth_token:
            apc = MSSQLProfileConfiguration
            profile = dict()

            profile[apc.ACCESS_TOKEN] = auth_token['accessToken']

            # TODO: When AAD/AD is supported
            """
            profile[apc.EXPIRES_ON] = auth_token['expiresOn']            
            profile[apc.REFRESH_TOKEN] = auth_token['refreshToken']
            profile[apc.TOKEN_TYPE] = auth_token['tokenType']
            """
            profile[apc.CONTROLLER_USER_NAME] = controller_username
            profile[apc.CLUSTER_NAME] = cluster_name
            profile[apc.ACCEPT_EULA] = accept_eula
            profile[apc.CONTROLLER_ENDPOINT] = \
                Configuration().DEFAULT_HOST if not controller_endpoint \
                else controller_endpoint

            # -- persist --
            self._profile_cache.set_all(profile)

    @property
    def access_token(self):
        """
        Gets the authenticated user's access-token.
        """
        return self._profile_cache.get(MSSQLProfileConfiguration.ACCESS_TOKEN)

    @property
    def refresh_token(self):
        """
        Gets the authenticated user's refresh-token
        """
        return self._profile_cache.get(MSSQLProfileConfiguration.REFRESH_TOKEN)

    @property
    def controller_endpoint(self):
        """
        Gets the SQL Server BDC host:port endpoint.
        """
        return self._profile_cache.get(MSSQLProfileConfiguration.CONTROLLER_ENDPOINT)

    @property
    def cluster_name(self):
        """
        Gets the SQL Server BDC name.
        """
        return self._profile_cache.get(MSSQLProfileConfiguration.CLUSTER_NAME)
    
    @property
    def accept_eula(self):
        """
        Gets the EULA acceptance.
        """
        return self._profile_cache.get(MSSQLProfileConfiguration.ACCEPT_EULA)

    @accept_eula.setter
    def accept_eula(self, a):
        self._profile_cache.set(MSSQLProfileConfiguration.ACCEPT_EULA, a)

    def get_current_account_user(self):
        profile = self._profile_cache.list()
        return profile.get(MSSQLProfileConfiguration.CONTROLLER_USER_NAME)

    def logout(self, user):
        """
        Removes a user's profile. There is no `logout` API call on MLServer.
        """
        if user is None:
            return

        profile = self._profile_cache.list()

        if profile.get(MSSQLProfileConfiguration.CONTROLLER_USER_NAME) == user:
                self._profile_cache.delete_all()

    def list(self):
        """
        Get the content of the profile configuration.
        """
        return self._profile_cache.list()

# -----------------------------------------------------------------------------#
# -----------------------------------------------------------------------------#
# -----------------------------------------------------------------------------#


class MSSQLProfileConfiguration(object):
    """
    Manages the `profile.json` cli profile configuration file. This includes
    persisting, deleting, and property lookups.
    """

    TOKEN_FIELDS_EXCLUDED_FROM_PERSISTENCE = ['tenantId']
    CONTROLLER_USER_NAME = 'controllerUserName'
    CONTROLLER_ENDPOINT = 'controllerEndpoint'
    ACCESS_TOKEN = 'accessToken'
    REFRESH_TOKEN = 'refreshToken'
    EXPIRES_IN = 'expiresIn'
    EXPIRES_ON = 'expiresOn'
    TOKEN_TYPE = 'tokenType'
    CLUSTER_NAME = 'clusterName'
    ACCEPT_EULA = 'acceptEula'

    def __init__(self, async_persist=None):
        # -- user home --
        cli_name = Configuration().CLI_NAME
        env_profile_override = '{}_CLI_PROFILE'.format(cli_name.upper())
        config_dir = os.path.expanduser(Configuration().CLI_CONFIG_DIR)
        ensure_dir(config_dir)

        self._profile = (os.environ.get(env_profile_override, None)
                         or os.path.join(config_dir,
                                         '{}-profile.json'.format(cli_name)))

        log.debug(os.environ.get(env_profile_override, None))
        log.debug('profile base location: %s', config_dir)
        log.debug('profile location: %s', self._profile)

        self._uncommitted_items = {}
        self._cached_profile = None
        self._should_flush_to_disk = False
        self._async_persist = async_persist

        if async_persist:
            import atexit
            atexit.register(self.flush_to_disk)

    def persist(self):
        """
        Persist any uncommitted items to the profile.
        """
        self._should_flush_to_disk = True
        if not self._async_persist:
            self.flush_to_disk()

    def flush_to_disk(self):
        # -- flush to disk --
        if self._should_flush_to_disk:
            with os.fdopen(os.open(self._profile,
                                   os.O_RDWR | os.O_CREAT | os.O_TRUNC, 0o600),
                           'w+') as pfile:

                # -- get profile from disk or cache--
                profile = self.list()

                # -- merge in updated `items` for override --
                new_profile = profile.copy()
                new_profile.update(self._uncommitted_items)

                # -- find empty items and remove from new profile --
                empty = {k: v for k, v in new_profile.items() if v is None}
                for k in empty:
                    del new_profile[k]

                # -- write to file --
                pfile.write(json.dumps(new_profile, indent=3))
                self._cached_profile = new_profile
                self._should_flush_to_disk = False

    @property
    def cache(self):
        return self.load_cache()

    def load_cache(self):
        if self._cached_profile is None:
            profile = {}

            # -- load file contents --
            if os.path.isfile(self._profile):
                try:
                    profile = FileUtil.read_json(self._profile) or {}
                except (CliError, ValueError) as ex:
                    raise CliError(
                        "Failed to load token files. Run '{0} login'. "
                        "(Inner Error: {1})".format(Configuration().CLI_NAME, ex))

            self._cached_profile = profile

        return self._cached_profile

    def list(self):
        """
        Lists all the key-value pairs in the profile configuration.
        :return:
        """

        self.load_cache()

        return self._cached_profile

    def delete(self, key):
        """
        Deletes the specified `key` in the profile configuration.
        :param key: Key to delete, in the profile configuration.
        :return:
        """
        item = dict()
        item[key] = None
        self._merge(item)
        self.persist()

    def get(self, key):
        """
        Gets the specified `key` in the profile configuration.

        :param key: Key to get in the profile configuration.
        :return: The values of the `key`.
        """

        profile = self.list()

        return profile.get(key)

    def set(self, key, value):
        """
        Sets the specified <key> <value> pair in the profile configuration.

        :param key: Key to set in the profile configuration.
        :param value: Value to set the `key` to.
        """
        item = dict()
        item[key] = value
        self._merge(item)
        self.persist()

    def set_all(self, items):
        self._merge(items)
        self.persist()

    def delete_all(self):
        """
        Removes the entire profile file and all its content.
        """
        try:
            self._uncommitted_items = {}
            self._cached_profile = None
            self._should_flush_to_disk = False

            # we can clear file contents, but deleting it is simpler
            os.remove(self._profile)
        except OSError as e:
            if e.errno != errno.ENOENT:
                raise

    def _merge(self, items=dict()):
        """
        Adds items to the uncommited map.
        """
        new_items = self._uncommitted_items.copy()
        new_items.update(items)
        self._uncommitted_items = new_items


_GLOBAL_MSSQL_PROFILE_CACHE = MSSQLProfileConfiguration(async_persist=True)
